/* Derived from Adafruit RGB_matrix_Panel library */

#include <Adafruit_GFX.h>   // Core graphics library
#include <P3RGB64x32MatrixPanel.h>
#include "ESP32_SD_UTF8toSJIS.h"
#include "ESP32_SD_ShinonomeFNT.h"
#include <WiFi.h>
#include <Wire.h>

// constructor with default pin wiring
// P3RGB64x32MatrixPanel matrix;
// use this constructor for custom pin wiring instead of the default above
// these pins are an example, you may modify this according to your needs
// P3RGB64x32MatrixPanel matrix(25, 26, 27, 21, 22, 23, 15, 32, 33, 12, 16, 17, 18); default
// P3RGB64x32MatrixPanel matrix(R1, G1, B1, R2, G2, B2, CLK, LAT, OE, A, B, C, D); default

P3RGB64x32MatrixPanel matrix(25, 26, 27, 21, 22, 13, 15, 32, 33, 12, 16, 17, 14);

#define TZ (9*60*60) /*JST*/

// POSTで読み取るデータサイズを定義
#define BUFSIZE 2048
// BOARDに表示可能な文字数を定義
#define BOARDWIDTH 100
// BOARDに表示するデコード済みのテキストを格納
char* decodeWordValue;
char* decodeColorValue;

WiFiServer server(80);

#define SSID "**********"   // 無線LANアクセスポイントのSSID
#define PASS "**********"       // パスワード

String lip;
const char* UTF8SJIS_file = "/font/Utf8Sjis.tbl"; //UTF8 Shift_JIS 変換テーブルファイル名を記載しておく
const char* Shino_Zen_Font_file = "/font/shnmk16.bdf"; //全角フォントファイル名を定義
const char* Shino_Half_Font_file = "/font/shnm8x16.bdf"; //半角フォントファイル名を定義
const uint8_t CS_SD = 5; //SD card CS ( Chip Select )
const uint8_t OE_LED= 33; //LED panel Output Enable
const uint8_t dTime=20;
const uint8_t msgL=4;
const String SPACE="　　　　";
String msg[10]={SPACE};
String xmsg[10]={SPACE};
String postMsg,postColor;

uint8_t col[40]={7};
uint8_t xcol[40]={7}; 
uint8_t rcol[40]={7};

uint8_t msgN,xmsgN,colN,rcolN;
uint16_t sj_length; //Shift_JISコードの長さ
uint8_t font_buf[32][16]; //東雲フォントデータ
uint8_t dot[128][16]={0};
uint8_t r=1;
uint16_t ptr=0;
uint16_t msgPtr=0;
uint16_t cn=0;
uint16_t cnt=0;
uint16_t x,y,i,cl,xb,len;
// 0,1,2,3,4,5,6,7,8,9,:,-,.,=,I,P
String alpha[16]=
  {"01110101010101010111",
   "00100110001000100111",
   "01110001011101000111",
   "01110001011100010111",
   "01010101010101110001",
   "01110100011100010111",
   "01110100011101010111",
   "01110001001000100010",
   "01110101011101010111",
   "01110101011100010111",
   "00000010000000100000",
   "00000000011100000000",
   "00000000000000000010",
   "00000111000001110000",
   "01110010001000100111",
   "01110101011101000100" };
ESP32_SD_ShinonomeFNT SFR(CS_SD, 24000000);

void setup() {
  Serial.begin(115200);
  WiFi.mode(WIFI_STA);                    // 無線LANをSTAモードに設定
  WiFi.begin(SSID,PASS);                  // 無線LANアクセスポイントへ接続
  while(WiFi.status() != WL_CONNECTED){   // 接続に成功するまで待つ
    delay(500);                           // 待ち時間処理
  }
  lip=WiFi.localIP().toString();
  
  configTime(TZ, 0, "ntp.nict.jp", "ntp.jst.mfeed.ad.jp"); // enable NTP
  
  matrix.begin();
  String lips[4]={"\0"};
  int count=split(lip,'.',lips);
  String lip1="";
  for (i=0; i<4; i++){
    lip1=lip1+add0(String(lips[i]),3);
    if (i<3) lip1=lip1+".";
  }
  int len=lip1.length();
  for (i=0; i<len; i++){
    String s=lip1.substring(i,i+1);
    if (s==".") drawAlpha(i*4+2,26,12); else drawAlpha(i*4+2,26,s.toInt()); 
  }
  
  delay(3000);
  matrix.setTextColor(matrix.color444(15,15,15));
  
  String msgs="残暑お見舞い申し上げます。";
  String mcols="23456";
  reformMessage(msgs,mcols);
  msgN=xmsgN;msgPtr=0;
  for (i=0; i<msgN; i++) msg[i]=xmsg[i];
  for (i=0; i<msgN*msgL; i++) col[i]=xcol[i];

  matrix.fillRect(0,8,64,16,matrix.color444(0, 0, 0));
  ptr=0;msgPtr=0; cnt=0;
  readMessage(0);
  readMessage(64);
}

void loop() {
  time_t t;
  static time_t last_t;
  struct tm *tm;
  t = time(NULL);

  tm = localtime(&t);
  String year1=String(tm->tm_year+1900);
  String month1=add0(String(tm->tm_mon+1),2);
  String day1=add0(String(tm->tm_mday),2);
  String hour1=add0(String(tm->tm_hour),2);
  String min1=add0(String(tm->tm_min),2);

  String date1=year1+"-"+month1+"-"+day1+"-"+hour1+":"+min1;
  len=date1.length();
  for (int i=0; i<len; i++){
    String s=date1.substring(i,i+1);
    if (s==":") drawAlpha(i*4,1,10);
    else if(s=="-") drawAlpha(i*4,1,11);
    else drawAlpha(i*4,1,s.toInt());    
  }

  
  getPOST();
  dispMessage(ptr); delay(dTime);
  ptr++; if (ptr==128) ptr=0;
  if (ptr==0 || ptr==64) readMessage(64-ptr);
}

void setPixel(uint8_t x,uint8_t y,uint8_t c){
  switch(c){
    case 0: matrix.drawPixel(x,y,matrix.color444(0,0,0)); break; // black
    case 1: matrix.drawPixel(x,y,matrix.color444(15,15,15)); break; // white
    case 2: matrix.drawPixel(x,y,matrix.color444(15,0,0)); break; // red
    case 3: matrix.drawPixel(x,y,matrix.color444(0,15,0)); break; // green
    case 4: matrix.drawPixel(x,y,matrix.color444(0,15,15)); break; // cyan
    case 5: matrix.drawPixel(x,y,matrix.color444(15,0,15)); break; // magenta
    case 6: matrix.drawPixel(x,y,matrix.color444(15,15,0)); break; // yellow
    case 7: matrix.drawPixel(x,y,matrix.color444(0,0,15)); break; // blue
  }
}

int split(String data, char delimiter, String *dst){
    int index = 0;
    int arraySize = (sizeof(data)/sizeof((data)[0]));  
    int datalength = data.length();
    for (int i = 0; i < datalength; i++) {
        char tmp = data.charAt(i);
        if ( tmp == delimiter ) {
            index++;
            if ( index > (arraySize - 1)) return -1;
        }
        else dst[index] += tmp;
    }
    return (index + 1);
}
void readMessage(uint8_t p){
  SFR.SD_Shinonome_Init3F(UTF8SJIS_file, Shino_Half_Font_file, Shino_Zen_Font_file);
  int dot1;
  sj_length = SFR.StrDirect_ShinoFNT_readALL(msg[msgPtr], font_buf);
  for (i=0; i<msgL;i++){
    for(y=0; y<16; y++){
      for (x=0; x<8; x++){
        dot1=bitRead(font_buf[2*i][y],7-x);
        dot[x+p+i*16][y]=dot1*10+col[cnt];
        Serial.print(dot1);
      }
      for (x=8; x<16; x++){
        dot1=bitRead(font_buf[2*i+1][y],15-x);
        dot[x+p+i*16][y]=dot1*10+col[cnt];
        Serial.print(dot1);       
      }
      Serial.println();
    }
    Serial.println("---------");
    cnt++; if(cnt==msgN*msgL) cnt=0;
  }
  SFR.SD_Shinonome_Close3F(); //StrDirect_ShinoFNT_readALL関数を使う場合、必ず閉じておく
  msgPtr++; if (msgPtr==msgN) msgPtr=0;
}

void dispMessage(uint8_t p){
  i=0;
  for (x=p; x<p+64; x++){
    xb=x; if (x>127) xb=x-128;
    for (y=0; y<16; y++){
      cl=dot[xb][y] % 10;
      if (dot[xb][y]>=10) setPixel(i,y+8,cl);
      else setPixel(i,y+8,0);
    }
    i++; 
  }
  delay(dTime);
}

void reformMessage(String mstr, String cstr){
  String mstr1=convToZen(mstr);
  mstr1=mstr1+"　　　　";
  int n=zenLength(mstr1);
  xmsgN=n/4;
  if (xmsgN>10) xmsgN=10;
  Serial.print("Strings:"); Serial.println(xmsgN);
  for (i=0; i<xmsgN; i++){
    xmsg[i]=mstr1.substring(i*12,i*12+12);
    Serial.println(xmsg[i]);
  }
  rcolN=cstr.length();
  Serial.print("Colors:"); Serial.println(rcolN);
  for (i=0; i<rcolN; i++){
    String c1=cstr.substring(i,i+1);
    rcol[i]=c1.toInt();
  }
  int cn=0;
  for (i=0; i < xmsgN*msgL; i++){
    xcol[i]=rcol[cn];
    Serial.print(xcol[i]); Serial.print(",");
    cn++;
    if (cn==rcolN) cn=0;
  }
  Serial.println();
}

void updateMessage(){
  if (xmsgN>0){
    msgN=xmsgN;
    for (i=0; i<msgN; i++) msg[i]=xmsg[i];
    for (i=0; i<msgN*msgL; i++) col[i]=xcol[i];             
    ptr=0; msgPtr=0; cnt=0;
    msgPtr=0; cnt=0;
    dot[128][16]={0};
    readMessage(0);
    readMessage(64);
  }    
}

void getPOST(){
  server.begin();
  // クライアントオブジェクトを取得
  WiFiClient client = server.available();
  // クライアントが存在するとき、clientはtrue
  if (client) {
    //Serial.println("new client");
    // クライアントからPOSTメソッドで呼ばれたらtrue
    boolean isPOST = false;
    // リクエストヘッダに改行（\r）以外が含まれたらfalse
    boolean currentLineIsBlank = true;
    // リクエストヘッダの終了（\r\n\r\n）を検出したらtrue
    boolean reqHeaderIsEnd = false;
    // リクエストデータを格納
    // 最大サイズはBUFSIZEで規定
    char reqParam[BUFSIZE];
    // リクエストデータを読むときに使用するインデックス
    long idx = 0;
    // POSTで送信されたテキストを格納
    char* wordValue;
    char* colorValue;
    // クライアントが存在する間、ぐるぐるまわる
    while (client.connected()) {
      // クライアントから受信するデータが存在するときはtrueを返す
      if (client.available()) {
        char c = client.read();
        //Serial.write(c);
        // （\r）\n（\r）\nのパターンがあらわれた場合、リクエストヘッダの終わりと見なしてreqHeaderIsEndをtrueにする
        if (c == '\n' && currentLineIsBlank) {
          reqHeaderIsEnd = true;
        }
        if (c == '\n') {
          currentLineIsBlank = true;

          // 行末にきたらidxをリセット
          idx = 0;
          // 行頭が"POST "ではじまっていたら、isPOSTをtrueにする
          if (strncmp(reqParam, "POST ", 5) == 0) {
            isPOST = true;
          }
        } 
        else if (c != '\r') {
          currentLineIsBlank = false;
          // リクエストデータ1行分をreqParamに格納する
          // 1行のデータサイズがPOSTデータの規定最大サイズBUFSIZE-1（-1は'\0'付加を考慮した分）を超える場合は、データを破棄
          if (idx < BUFSIZE - 1) {
            reqParam[idx] = c;
            idx++;
          }
        }
      }
      // クライアントから受信するデータがなくなったらループを抜ける
      else {
        break;
      }
    }
    // POSTで呼び出された場合のデータ解析とデコード処理
    if (isPOST) {
      // POSTデータを'&'で分割したパラメタを格納
      char* splitReqParam;
      // POSTデータをstrtok_rでパラメタに分割するときに使用
      char* reqParamPtr;
      // パラメタを'='で分割した左辺（パラメタ名）を格納
      char* paramName;
      // パラメタを'='で分割した右辺（パラメタ値）を格納
      char* paramValue;
      // パラメタをstrtok_rで名前と値に分割するときに使用
      char* nameValuePtr;
      // reqParamに格納した文字列の後ろにNULLを追加して、文字列の終わりを示す
      reqParam[idx] = '\0';

      // POSTデータを'&'で分割してsplitReqParamに格納
      for (splitReqParam = strtok_r(reqParam, "&", &reqParamPtr); splitReqParam != NULL; splitReqParam = strtok_r(NULL, "&", &reqParamPtr)) {
        // パラメタsplitReqParamを'='で分割して、名前をparamName、値をparamValueに格納
        // ネストしたいので、strtokの代わりにstrtok_rを使用
        char* paramName = strtok_r(splitReqParam, "=", &nameValuePtr);
        char* paramValue = strtok_r(NULL, "=", &nameValuePtr);

        // パラメタ名が"word"であれば、テキストとみなしてwordValueに値を格納
        if (strcmp(paramName, "word") == 0) {
          wordValue = paramValue;
        }
        // パラメタ名が"color"であれば、テキストとみなしてwordValueに値を格納
        if (strcmp(paramName, "color") == 0) {
          colorValue = paramValue;
        }
      }

      // POSTで送信されたデータはパーセントエンコーディングされているため、デコードする関数を呼ぶ
      // デコードされた文字列はdecodeWordValueに格納
      decodeWordValue = strdup(wordValue);
      urlDecode(wordValue, decodeWordValue);
      // POSTで送信されたデータはパーセントエンコーディングされているため、デコードする関数を呼ぶ
      // デコードされた文字列はdecodeWordValueに格納
      decodeColorValue = strdup(colorValue);
      urlDecode(colorValue, decodeColorValue);
    }
    // クライアントにデータを返す
    client.println("HTTP/1.1 200 OK");
    client.println("Content-Type: text/html");
    client.println("Connnection: close");
    client.println();
    client.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    client.println("<!DOCTYPE html>");
    client.println("<html lang=\"ja\">");
    client.println("<head>");
    client.println("<meta http-equiv=\"content-type\" content=\"text/html\" />");
    client.println("<title>");
    client.println("電光掲示板");
    client.println("</title>");
    client.println("</head>");
    client.println("<body>");
    client.println("<font size='3'>");
    client.println("電光掲示板<br><br>");
    client.println("<form method=\"post\">");
    client.println("送信文字列:");
    client.println("<input type=\"text\" name=\"word\" value='' size=40"); if (isPOST) { client.print(decodeWordValue); } client.print("\" />");
    client.println("<br>");
    client.println("文字色設定:");
    client.println("<input type=\"text\" name=\"color\" value='' size=40"); if (isPOST) { client.print(decodeColorValue); } client.print("\" />");
    client.println("<br>");
    client.println("<input type=\"submit\" value=\"SEND\" />");  
    client.println("<input type=\"reset\" value=\"CLEAR\" />");
    client.println("</form>");
    client.println("</font>");
    client.println("</body>");
    client.println("</html>");
    // クライアントをサーバから切断
    delay(100);
    client.stop();
    //Serial.println("client disconnected");
    // POSTで呼び出された場合
    if (isPOST) {  
      // dispMsgはBOARDに表示する文字列を格納する変数
      // 変数のサイズはBOARDに表示可能な文字数+1（'\0'を考慮して1文字分多く確保）
      char dispMsg1[BOARDWIDTH + 1];
      char dispMsg2[BOARDWIDTH + 1];
      // テキストからBOARDに表示可能な文字数を切り取ってdispMsgに格納
      strncpy(dispMsg1, decodeWordValue, BOARDWIDTH);
      // strncpyは'\0'をつけてくれないので、自分でつける
      dispMsg1[BOARDWIDTH] = '\0';
      strncpy(dispMsg2, decodeColorValue, BOARDWIDTH);
      // strncpyは'\0'をつけてくれないので、自分でつける
      dispMsg2[BOARDWIDTH] = '\0';
      postMsg=dispMsg1;
      postColor=dispMsg2;
      Serial.print("post message: ");Serial.println(postMsg);
      Serial.print("post colors: ");Serial.println(postColor);
      delay(100);
      reformMessage(postMsg,postColor);
      delay(100);
      updateMessage();
    }
  }
}

boolean urlDecode(char* str, char* dstStr) {
  // POSTで読み取ったデータはパーセントエンコーディングされているため、デコードを実施
  // char* str : 変換する文字列を格納、 char* dstStr : 変換後の文字列を格納
  if (str != NULL) {
    long dstIdx = 0;
    for(long idx = 0; idx < strlen(str); idx++) {
      if (str[idx] == '%') {
        char tmpHex[] = { str[idx+1], str[idx+2] };
        dstStr[dstIdx] = strtol(tmpHex, NULL, 16);
        idx += 2;
      }
      else if (str[idx] == '+') {
        dstStr[dstIdx] = ' ';
      }
      else {
        dstStr[dstIdx] = str[idx];
      }
      dstIdx++;
    }
    dstStr[dstIdx] = '\0';
  }
  return true;
}

int zenLength(String s){
  return s.length()/3;
}

String convToZen (String s1){
  char chr1[601]={0};
  char chr2[601]={0};
  int ln=s1.length();
  if (ln>600) ln=600;
  s1.toCharArray(chr1,ln+1);
  int ptr1=0;
  int ptr2=0;
  String s2;
  
  while (ptr1<ln){
    if (chr1[ptr1]>=0xE0 && chr1[ptr1]<=0xEF 
     && chr1[ptr1+1]>=0x80 && chr1[ptr1+1]<=0xBF
     && chr1[ptr1+2]>=0x80 && chr1[ptr1+2]<=0xBF){
      chr2[ptr2]=chr1[ptr1];
      chr2[ptr2+1]=chr1[ptr1+1];
      chr2[ptr2+2]=chr1[ptr1+2];
      ptr1=ptr1+3;
      ptr2=ptr2+3;
    } else if (chr1[ptr1]==0x20) {
      chr2[ptr2]=0xE3;
      chr2[ptr2+1]=0x80;
      chr2[ptr2+2]=0x80;
      ptr1++;
      ptr2=ptr2+3; 
    } else if (chr1[ptr1]>0x20 && chr1[ptr1]<0x60){
      chr2[ptr2]=0xEF;
      chr2[ptr2+1]=0xBC;
      chr2[ptr2+2]=chr1[ptr1]+0x60;
      ptr1++;
      ptr2=ptr2+3;
    } else if (chr1[ptr1]>=0x60 && chr1[ptr1]<0x7F){
      chr2[ptr2]=0xEF;
      chr2[ptr2+1]=0xBD;
      chr2[ptr2+2]=chr1[ptr1]+0x20;
      ptr1++;
      ptr2=ptr2+3;      
    } else {
      ptr1++;
    }
  }
  s2=chr2;
  return s2;
}

void drawAlpha(uint8_t x, uint8_t y, uint8_t n){
  
  for (int j=0; j<5; j++){
    for (int i=0; i<4; i++){
      int k=j*4+i;
      if (alpha[n].substring(k,k+1)=="1") setPixel(x+i,y+j,1); else setPixel(x+i,y+j,0);
    }
  }
}

String add0(String s,int n){
  s="0000"+s;
  int len=s.length();
  return s.substring(len-n,len);
}
